from random import randint, random, uniform, choice
from .problem import Problem
from .algorithm_genetic import GeneralEvolutionaryAlgorithm
from .operators import RandomGenerator, PmMutator, ParetoDominance, EpsilonDominance, crowding_distance, \
    NonUniformMutation, UniformMutator, CopySelector
from .archive import Archive
from copy import copy
import time
import math


class SwarmAlgorithm(GeneralEvolutionaryAlgorithm):
    def __init__(self, problem: Problem, name="General Swarm-based Algorithm"):
        super().__init__(problem, name)
        # self.options.declare(name='v_max', default=, lower=0., desc='maximum_allowed_speed')
        self.global_best = None  # swarm algorithms share the information, who is the leader
        self.dominance = ParetoDominance()  # some dominance should be defined for every kind of multi-opbjective swarm

    def init_pvelocity(self, population):
        pass

    def init_pbest(self, population):
        for individual in population:
            individual.features['best_cost'] = individual.costs_signed
            individual.features['best_vector'] = individual.vector

    def khi(self, c1: float, c2: float) -> float:
        """
        Constriction coefficient [1].
        [1] Ebarhart and Kennedym Empirical study of particle swarm optimization,” in Proc. IEEE Int. Congr.
        Evolutionary Computation, vol. 3, 1999, pp. 101–106.

        :param c1: specific parameter to control the particle best component.
        :param c2: specific parameter to control the global best component.
        :return: float, constriction coefficient
        """
        rho = c1 + c2
        if rho <= 4:
            result = 1.0
        else:
            result = 2.0 / (2.0 - rho - (rho ** 2.0 - 4.0 * rho) ** 0.5)

        return result

    def speed_constriction(self, velocity, u_bound, l_bound) -> float:
        """
        Velocity constriction factor [1].

        .. Ref:
        [1] Nebro, Antonio J., et al. "SMPSO: A new PSO-based metaheuristic for multi-objective optimization."
            2009 IEEE Symposium on Computational Intelligence in Multi-Criteria Decision-Making (MCDM). IEEE, 2009.

        :param velocity: parameter velocity for the i^th component
        :param ub: upper bound
        :param lb: lower bound
        :return:
        """

        delta_i = (u_bound - l_bound) / 2.
        # user defined max speed
        velocity = min(velocity, delta_i)
        velocity = max(velocity, -delta_i)

        return velocity

    def select_leader(self):
        pass

    def inertia_weight(self):
        pass

    def update_global_best(self, offsprings):
        pass

    def update_velocity(self, population):
        pass

    def update_position(self, population):
        pass

    def update_particle_best(self, population):
        for particle in population:
            flag = self.dominance.compare(particle.costs_signed, particle.features['best_cost'])
            if flag != 2:
                particle.features['best_cost'] = particle.costs_signed
                particle.features['best_vector'] = particle.vector

    def turbulence(self, particles, current_step=0):
        pass

    def run(self):
        pass


class OMOPSO(SwarmAlgorithm):
    """
    Implementation of OMOPSO, a multi-objective particle swarm optimizer (MOPSO).
    OMOPSO uses Crowding distance, Mutation and ε-Dominance.
    According to [3], OMOPSO is one of the top-performing PSO algorithms.

    [1] Margarita Reyes SierraCarlos A. Coello Coello
        Improving PSO-Based Multi-objective Optimization Using Crowding, Mutation and ∈-Dominance
        DOI https://doi.org/10.1007/978-3-540-31880-4_35
    [2] S. Mostaghim ; J. Teich :
        Strategies for finding good local guides in multi-objective particle swarm optimization (MOPSO)
        DOI: 10.1109/SIS.2003.1202243
    [3] Durillo, J. J., J. Garcia-Nieto, A. J. Nebro, C. A. Coello Coello, F. Luna, and E. Alba (2009).
        Multi-Objective Particle Swarm Optimizers: An Experimental Comparison.
        Evolutionary Multi-Criterion Optimization, pp. 495-509
    """

    def __init__(self, problem: Problem, name="OMOPSO"):
        super().__init__(problem, name)
        self.options.declare(name='prob_mutation', default=0.1, lower=0,
                             desc='prob_mutation'),
        self.options.declare(name='epsilons', default=0.01, lower=1e-6,
                             desc='prob_epsilons')
        self.n = self.options['max_population_size']

        self.selector = CopySelector(self.problem.parameters)
        self.dominance = ParetoDominance()

        self.individual_features['velocity'] = dict()
        self.individual_features['best_cost'] = dict()
        self.individual_features['best_vector'] = dict()

        # set random generator
        self.generator = RandomGenerator(self.problem.parameters, self.individual_features)
        self.leaders = Archive()
        self.archive = Archive(dominance=EpsilonDominance(epsilons=self.options['epsilons']))

        self.non_uniform_mutator = NonUniformMutation(self.problem.parameters, self.options['prob_mutation'],
                                                      self.options['max_population_number'])
        self.uniform_mutator = UniformMutator(self.problem.parameters, self.options['prob_mutation'],
                                              self.options['max_population_number'])
        # constants for the speed and the position calculation
        self.c1_min = 1.5
        self.c1_max = 2.0
        self.c2_min = 1.5
        self.c2_max = 2.0
        self.r1_min = 0.0
        self.r1_max = 1.0
        self.r2_min = 0.0
        self.r2_max = 1.0
        self.min_weight = 0.1
        self.max_weight = 0.5

    def inertia_weight(self):
        return uniform(self.min_weight, self.max_weight)

    def init_pvelocity(self, population):
        """
        Inits the particle velocity and its allowed maximum speed.
        :param population: list of individuals
        :return
        """
        for individual in population:
            # the initial speed is set to zero
            individual.features['velocity'] = [0] * len(individual.vector)

        return

    def turbulence(self, particles, current_step=0):
        """
        OMOPSO applies a combination of uniform and nonuniform
        mutation to the particle swarm(uniform mutation to the first 30 % of
        the swarm, non - uniform to the next 30 %, and no mutation on the particles)
        """

        for i in range(len(particles)):
            if i % 3 == 0:
                mutated = self.uniform_mutator.mutate(particles[i])
            elif i % 3 == 1:
                mutated = self.non_uniform_mutator.mutate(particles[i], current_step)
            particles[i].vector = copy(mutated.vector)
        return

    def update_velocity(self, individuals):
        for individual in individuals:
            individual.features['velocity'] = [0] * len(individual.vector)
            global_best = self.select_leader()

            r1 = round(uniform(self.r1_min, self.r1_max), 1)
            r2 = round(uniform(self.r2_min, self.r2_max), 1)
            c1 = round(uniform(self.c1_min, self.c1_max), 1)
            c2 = round(uniform(self.c2_min, self.c2_max), 1)

            for i in range(0, len(individual.vector)):
                momentum = self.inertia_weight() * individual.vector[i]
                v_cog = c1 * r1 * (individual.features['best_vector'][i] - individual.vector[i])
                v_soc = c2 * r2 * (global_best.vector[i] - individual.vector[i])

                v = self.khi(c1, c2) * (momentum + v_cog + v_soc)
                individual.features['velocity'][i] = self.speed_constriction(v, self.parameters[i]['bounds'][1],
                                                                             self.parameters[i]['bounds'][0])

    def update_position(self, individuals):
        for individual in individuals:
            for parameter, i in zip(self.parameters, range(len(individual.vector))):
                individual.vector[i] = individual.vector[i] + individual.features['velocity'][i]

                # adjust maximum position if necessary
                if individual.vector[i] > parameter['bounds'][1]:
                    individual.vector[i] = parameter['bounds'][1]
                    individual.features['velocity'][i] *= -1

                # adjust minimum position if necessary
                if individual.vector[i] < parameter['bounds'][0]:
                    individual.vector[i] = parameter['bounds'][0]
                    individual.features['velocity'][i] *= -1

    def update_global_best(self, swarm):
        """ Manages the leader class in OMOPSO. """

        # the fitness of the particles are calculated by their crowding distance
        crowding_distance(swarm)

        # the length of the leaders archive cannot be longer than the number of the initial population
        self.leaders += swarm
        self.leaders.truncate(self.options['max_population_size'], 'crowding_distance')
        self.archive += swarm

        return

    def select_leader(self):
        """
        There are different possibilities to select the global best solution.
        The leader class in this concept contains everybody after the initialization, every individual expected as a
        leader, we select 2 from them and select the non-dominated as the global best.

        :return:
        """

        if self.leaders.size() == 1:
            return self.leaders.rand_choice()

        candidates = self.leaders.rand_sample(2)

        # randomly favourize one of them
        # best_global = choice(candidates)

        # should select those which has bigger fitness
        # # if one of them dominates, it will be selected as global best
        # dom = self.dominance.compare(candidates[0].costs_signed, candidates[1].costs_signed)
        #
        # if dom == 1:
        #     best_global = candidates[0]
        #
        # if dom == 2:
        #     best_global = candidates[1]

        if candidates[1].features['crowding_distance'] > candidates[0].features['crowding_distance']:
            best_global = candidates[1]
        else:
            best_global = candidates[0]
        return best_global

    def run(self):
        t_s = time.time()
        self.problem.logger.info("PSO: {}/{}".format(self.options['max_population_number'],
                                                     self.options['max_population_size']))
        # update mutators
        self.non_uniform_mutator = NonUniformMutation(self.problem.parameters, self.options['prob_mutation'],
                                                      self.options['max_population_number'])
        self.uniform_mutator = UniformMutator(self.problem.parameters, self.options['prob_mutation'],
                                              self.options['max_population_number'])
        # initialize the swarm
        self.generator.init(self.options['max_population_size'])
        individuals = self.generator.generate()

        for individual in individuals:
            # append to problem
            self.problem.individuals.append(individual)
            # add to population
            individual.population_id = 0

        self.evaluate(individuals)

        self.init_pvelocity(individuals)
        self.init_pbest(individuals)
        self.update_global_best(individuals)

        # sync to datastore
        for individual in individuals:
            self.problem.data_store.sync_individual(individual)

        it = 0
        while it < self.options['max_population_number']:
            offsprings = self.selector.select(individuals)

            self.update_velocity(offsprings)
            self.update_position(offsprings)
            self.turbulence(offsprings, it)

            self.evaluate(offsprings)

            self.update_particle_best(offsprings)
            self.update_global_best(offsprings)

            # update individuals
            individuals = offsprings

            for individual in individuals:
                # add to population
                individual.population_id = it + 1
                # append to problem
                self.problem.individuals.append(individual)
                # sync to datastore
                self.problem.data_store.sync_individual(individual)

            it += 1

        t = time.time() - t_s
        self.problem.logger.info("PSO: elapsed time: {} s".format(t))

        # sync changed individual informations
        self.problem.data_store.sync_all()


class SMPSO(SwarmAlgorithm):
    """
    Implementation of SMPSP, a multi-objective particle swarm optimizer (MOPSO).
    OMOPSO uses Crowding distance, Mutation and ε-Dominance.
    According to [3], SMPSP is one of the top-performing PSO algorithms. There are 3 key-differences between OMOPS§ and
    SMPSO, the mutator is polynomial mutation, the moment component of the velocity is constant, the values of the C1
    and C2 values are [1.5, 2.5], while in the case of OMOPSO they are selected from [1.5, 2.0].
    Here, instead of reversing the values from the borders, their speed are reduced by multiplying it by
    0.001 [3].

    Both of the SMPSO and OMOPSO can be defined with and without epsilon-dominance archive.

    [1] Margarita Reyes SierraCarlos A. Coello Coello
        Improving PSO-Based Multi-objective Optimization Using Crowding, Mutation and ∈-Dominance
        DOI https://doi.org/10.1007/978-3-540-31880-4_35
    [2] S. Mostaghim ; J. Teich :
        Strategies for finding good local guides in multi-objective particle swarm optimization (MOPSO)
        DOI: 10.1109/SIS.2003.1202243
    [3] Durillo, J. J., J. Garcia-Nieto, A. J. Nebro, C. A. Coello Coello, F. Luna, and E. Alba (2009).
        Multi-Objective Particle Swarm Optimizers: An Experimental Comparison.
        Evolutionary Multi-Criterion Optimization, pp. 495-509
    """

    def __init__(self, problem: Problem, name="SMPSO Algorithm"):
        super().__init__(problem, name)
        self.options.declare(name='prob_mutation', default=0.1, lower=0,
                             desc='prob_mutation'),
        self.n = self.options['max_population_size']

        self.individual_features['velocity'] = dict()
        self.individual_features['best_cost'] = dict()
        self.individual_features['best_vector'] = dict()

        self.selector = CopySelector(self.problem.parameters)
        self.dominance = ParetoDominance()
        # set random generator
        self.generator = RandomGenerator(self.problem.parameters, self.individual_features)
        self.leaders = Archive()
        self.mutator = PmMutator(self.problem.parameters, self.options['prob_mutation'])
        # constants for the speed and the position calculation

        self.c1_min = 1.5
        self.c1_max = 2.5
        self.c2_min = 1.5
        self.c2_max = 2.5
        self.r1_min = 0.0
        self.r1_max = 1.0
        self.r2_min = 0.0
        self.r2_max = 1.0
        self.min_weight = 0.1
        self.max_weight = 0.1

    def inertia_weight(self):
        return uniform(self.min_weight, self.max_weight)

    def init_pvelocity(self, individuals):
        """
        Inits the particle velocity and its allowed maximum speed.
        :param population: list of individuals
        :return
        """
        for individual in individuals:
            # the initial speed is set to zero
            individual.features['velocity'] = [0] * len(individual.vector)

        return

    def turbulence(self, particles, current_step=0):
        """ SMPSO applies polynomial mutation on 15% of the particles """

        for i in range(len(particles)):
            if i % 6 == 0:
                mutated = self.mutator.mutate(particles[i])
                particles[i].vector = copy(mutated.vector)

    def update_velocity(self, individuals):
        for individual in individuals:
            individual.features['velocity'] = [0] * len(individual.vector)
            global_best = self.select_leader()

            r1 = round(uniform(self.r1_min, self.r1_max), 1)
            r2 = round(uniform(self.r2_min, self.r2_max), 1)
            c1 = round(uniform(self.c1_min, self.c1_max), 1)
            c2 = round(uniform(self.c2_min, self.c2_max), 1)

            for i in range(0, len(individual.vector)):
                momentum = self.inertia_weight() * individual.vector[i]
                v_cog = c1 * r1 * (individual.features['best_vector'][i] - individual.vector[i])
                v_soc = c2 * r2 * (global_best.vector[i] - individual.vector[i])

                v = self.khi(c1, c2) * (momentum + v_cog + v_soc)
                individual.features['velocity'][i] = self.speed_constriction(v, self.parameters[i]['bounds'][1],
                                                                             self.parameters[i]['bounds'][0])

    def update_position(self, individuals):
        for individual in individuals:
            for parameter, i in zip(self.parameters, range(len(individual.vector))):
                individual.vector[i] = individual.vector[i] + individual.features['velocity'][i]

                # adjust maximum position if necessary
                if individual.vector[i] > parameter['bounds'][1]:
                    individual.vector[i] = parameter['bounds'][1]
                    individual.features['velocity'][i] *= 0.001

                # adjust minimum position if necessary
                if individual.vector[i] < parameter['bounds'][0]:
                    individual.vector[i] = parameter['bounds'][0]
                    individual.features['velocity'][i] *= 0.001

    def update_global_best(self, swarm):
        """ Manages the leader class in OMOPSO. """

        # the fitness of the particles are calculated by their crowding distance
        crowding_distance(swarm)

        # the length of the leaders archive cannot be longer than the number of the initial population
        self.leaders += swarm
        self.leaders.truncate(self.options['max_population_size'], 'crowding_distance')
        #self.problem.archive += swarm

        return

    def select_leader(self):
        """
        There are different possibilities to select the global best solution.
        The leader class in this concept contains everybody after the initialization, every individual expected as a
        leader, we select 2 from them and select the non-dominated as the global best.

        :return:
        """

        if self.leaders.size() == 1:
            return self.leaders.rand_choice()

        candidates = self.leaders.rand_sample(2)

        # randomly favourize one of them
        # best_global = choice(candidates)

        # should select those which has bigger fitness
        # # if one of them dominates, it will be selected as global best
        # dom = self.dominance.compare(candidates[0].costs_signed, candidates[1].costs_signed)
        #
        # if dom == 1:
        #     best_global = candidates[0]
        #
        # if dom == 2:
        #     best_global = candidates[1]

        if candidates[1].features['crowding_distance'] > candidates[0].features['crowding_distance']:
            best_global = candidates[1]
        else:
            best_global = candidates[0]
        return best_global

    def run(self):
        t_s = time.time()
        self.problem.logger.info("PSO: {}/{}".format(self.options['max_population_number'],
                                                     self.options['max_population_size']))
        # initialize the swarm
        self.generator.init(self.options['max_population_size'])
        individuals = self.generator.generate()

        for individual in individuals:
            # append to problem
            self.problem.individuals.append(individual)
            # add to population
            individual.population_id = 0

        self.evaluate(individuals)

        self.init_pvelocity(individuals)
        self.init_pbest(individuals)
        self.update_global_best(individuals)

        # sync to datastore
        for individual in individuals:
            self.problem.data_store.sync_individual(individual)

        it = 0
        while it < self.options['max_population_number']:
            offsprings = self.selector.select(individuals)

            self.update_velocity(offsprings)
            self.update_position(offsprings)
            self.turbulence(offsprings, it)

            self.evaluate(offsprings)

            self.update_particle_best(offsprings)
            self.update_global_best(offsprings)

            # update individuals
            individuals = offsprings

            for individual in individuals:
                # add to population
                individual.population_id = it + 1
                # append to problem
                self.problem.individuals.append(individual)
                # sync to datastore
                self.problem.data_store.sync_individual(individual)

            it += 1

        t = time.time() - t_s
        self.problem.logger.info("PSO: elapsed time: {} s".format(t))

        # sync changed individual informations
        self.problem.data_store.sync_all()

# ........................
#
# ........................

# class PSO_V1(SwarmAlgorithm):
#     """
#
#     X. Li. A Non-dominated Sorting Particle Swarm Optimizer for Multiobjective
#     Optimization. In Genetic and Evolutionary Computation - GECCO 2003, volume
#     2723 of LNCS, pages 37–48, 2003.
#
#     This algorithm is a variant of the original PSO, published by Eberhart(2000), the origin of this modified algorithm,
#     which constriction factor was introduced by Clercs in 1999.
#
#     The code is based on SHI and EBERHARTS algorithm.
#
#     Empirical study of particle swarm optimization,” in Proc. IEEE Int. Congr. Evolutionary Computation, vol. 3,
#     1999, pp. 101–106.
#     """
#
#     def __init__(self, problem: Problem, name="Particle Swarm Algorithm - with time varieting inertia weight"):
#         super().__init__(problem, name)
#         self.n = self.options['max_population_size']
#         self.mutator = SwarmStepTVIW(self.problem.parameters, self.options['max_population_number'])
#         self.selector = DummySelector(self.problem.parameters, self.problem.signs)
#
#     def run(self):
#         # set random generator
#         self.generator = RandomGenerator(self.problem.parameters)
#         self.generator.init(self.options['max_population_size'])
#
#         population = self.gen_initial_population()
#         self.evaluate(population.individuals)
#         self.add_features(population.individuals)
#
#         for individual in population.individuals:
#             self.mutator.evaluate_best_individual(
#                 individual)  # TODO: all evaluating should be derived from Evaluator class
#
#         self.selector.fast_nondominated_sorting(population.individuals)
#         self.populations.append(population)
#
#         t_s = time.time()
#         self.problem.logger.info("PSO: {}/{}".format(self.options['max_population_number'],
#                                                      self.options['max_population_size']))
#
#         i = 0
#         while i < self.options['max_population_number']:
#             offsprings = self.selector.select(population.individuals)
#
#             pareto_front = []
#             for individual in offsprings:
#                 if individual.features['front_number'] == 1:
#                     pareto_front.append(individual)
#
#             for individual in offsprings:
#                 index = randint(0, len(pareto_front) - 1)  # takes random individual from Pareto front
#                 best_individual = pareto_front[index]
#                 if best_individual is not individual:
#                     self.mutator.update(best_individual)
#                     self.mutator.mutate(individual)
#
#             self.evaluate(offsprings)
#             self.add_features(offsprings)
#
#             for individual in offsprings:
#                 self.mutator.evaluate_best_individual(individual)
#
#             self.selector.fast_nondominated_sorting(offsprings)
#             population = Population(offsprings)
#             self.populations.append(population)
#
#             i += 1
#
#         t = time.time() - t_s
#         self.problem.logger.info("PSO: elapsed time: {} s".format(t))



import unittest
from unittest import TestCase, main

from ..executor import CondorComsolJobExecutor, CondorMatlabJobExecutor, CondorPythonJobExecutor, CondorCSTJobExecutor
from ..algorithm_genetic import NSGAII
from ..problem import Problem
from ..algorithm import DummyAlgorithm
from ..individual import Individual
from ..config import config

import random
import pathlib
import os
import zipfile


class CondorMatlabProblem(Problem):
    """ Describe simple one objective optimization problem. """

    def set(self):
        self.name = "CondorMatlabProblem"
        self.parameters = [{'name': 'a', 'initial_value': 10, 'bounds': [0, 20]},
                      {'name': 'b', 'initial_value': 10, 'bounds': [5, 15]}]
        self.costs = [{'name': 'F_1'}]
        file_path = os.path.join(str(pathlib.Path(__file__).parent.absolute()), "data/run_input.m")
        self.executor = CondorMatlabJobExecutor(self,
                                                script=file_path,
                                                parameter_file="input.txt",
                                                files_from_condor=["output.txt"])

    def evaluate(self, individual):
        return self.executor.eval(individual)

    def parse_results(self, output_files, individual):
        with open(output_files[0]) as file:
            content = file.readlines()
        return [float(content[0])]


class ComsolProblem(Problem):
    """ Describe simple one objective optimization problem. """

    def set(self):
        self.name = "CondorComsolProblem"
        self.parameters = [{'name': 'a', 'initial_value': 10, 'bounds': [0, 20]},
                           {'name': 'b', 'initial_value': 10, 'bounds': [5, 15]}]

        self.costs = [{'name': 'F_1'}]
        file_path = os.path.join(str(pathlib.Path(__file__).parent.absolute()), "data/elstat.mph")
        self.executor = CondorComsolJobExecutor(self, model_file=file_path,
                                                files_from_condor=["out.txt", "elstat.mph"])

    def evaluate(self, individual):
        return self.executor.eval(individual)

    def parse_results(self, output_files, individual):
        with open(output_files[0]) as file:
            content = file.readlines()
        line_with_results = content[5]  # 5th line contains results
        result = float(line_with_results)
        return [result]


class PythonExecProblem(Problem):
    """ Describe simple one objective optimization problem. """
    def set(self):
        self.parameters = [{'name': 'a', 'initial_value': 10, 'bounds': [0, 20]},
                           {'name': 'b', 'initial_value': 10, 'bounds': [5, 15]}]
        self.costs = [{'name': 'F_1'}]
        file_path = os.path.join(str(pathlib.Path(__file__).parent.absolute()), "data/run_exec.py")
        self.executor = CondorPythonJobExecutor(self,
                                                script=file_path,
                                                parameter_file=None,
                                                output_files=["output.txt"])

    def evaluate(self, individual):
        return self.executor.eval(individual)

    def parse_results(self, output_files, individual):
        with open(output_files[0]) as file:
            content = file.readlines()
        return [float(content[0])]


class PythonExecProblemNSGAII(Problem):
    """ Describe simple one objective optimization problem. """
    def set(self):
        self.parameters = [{'name': 'a', 'initial_value': 10, 'bounds': [0, 20]},
                           {'name': 'b', 'initial_value': 10, 'bounds': [5, 15]}]
        self.costs = [{'name': 'F_1'}]
        file_path = os.path.join(str(pathlib.Path(__file__).parent.absolute()), "data/run_exec.py")
        self.executor = CondorPythonJobExecutor(self,
                                                script=file_path,
                                                parameter_file=None,
                                                output_files=["output.txt"])

    def evaluate(self, individual):
        return self.executor.eval(individual)

    def parse_results(self, output_files, individual):
        with open(output_files[0]) as file:
            content = file.readlines()
        return [float(content[0])]


class PythonInputProblem(Problem):
    """ Describe simple one objective optimization problem. """
    def set(self):
        self.parameters = [{'name': 'a', 'initial_value': 10, 'bounds': [0, 20]},
                           {'name': 'b', 'initial_value': 10, 'bounds': [5, 15]}]
        self.costs = [{'name': 'F_1'}]
        file_path = os.path.join(str(pathlib.Path(__file__).parent.absolute()), "data/run_input.py")
        self.executor = CondorPythonJobExecutor(self,
                                                script=file_path,
                                                parameter_file="input.txt",
                                                output_files=["output.txt"])

    def evaluate(self, individual):
        return self.executor.eval(individual)

    def parse_results(self, output_files, individual):
        with open(output_files[0]) as file:
            content = file.readlines()
        return [float(content[0])]


class CSTProblem(Problem):
    """ Describe simple one objective optimization problem. """
    def set(self):
        self.parameters = [{'name': 'a', 'initial_value': 0.1, 'bounds': [0, 1]}]
        self.costs = [{'name': 'F_1'}]
        file_path = os.path.join(str(pathlib.Path(__file__).parent.absolute()), "data/elstat.cst")
        self.executor = CondorCSTJobExecutor(self,
                                             model_file=file_path,
                                             files_from_condor=["elstat.zip"])

    def evaluate(self, individual):
        return self.executor.eval(individual)

    def parse_results(self, output_files, individual):
        archive = zipfile.ZipFile(output_files[0], 'r')
        file = archive.open('elstat/Export/Es Solver/Energy.txt')
        content = file.readlines()

        return [float(content[0])]

    def evaluate(self, individual):
        return self.executor.eval(individual)


class TestCondor(TestCase):
    """ Tests simple optimization problem where calculation of
        goal function is submitted as a job on HtCondor.
    """

    @unittest.skipIf(config["condor_host"] is None, "Condor is not defined.")
    def test_condor_matlab_input(self):
        """ Tests one calculation of goal function."""
        problem = CondorMatlabProblem()

        individuals = [Individual([1, 2])]
        evaluator = DummyAlgorithm(problem)
        evaluator.evaluate(individuals)

        self.assertAlmostEqual(5, individuals[0].costs[0])

    @unittest.skipIf(config["condor_host"] is None, "Condor is not defined.")
    def test_condor_comsol_exec(self):
        """ Tests one calculation of goal function."""
        problem = ComsolProblem()

        individuals = [Individual([10, 10])]
        algorithm = DummyAlgorithm(problem)
        algorithm.evaluate(individuals)

        self.assertAlmostEqual(112.94090668383139, individuals[0].costs[0])

    @unittest.skipIf(config["condor_host"] is None, "Condor is not defined.")
    def test_condor_cst_exec(self):
        """ Tests one calculation of goal function."""
        problem = CSTProblem()

        individuals = [Individual([0.7])]
        algorithm = DummyAlgorithm(problem)
        algorithm.evaluate(individuals)

        # self.assertAlmostEqual(112.94090668383139, individuals[0].costs[0])
        self.assertAlmostEqual(0, 0)

    @unittest.skipIf(config["condor_host"] is None, "Condor is not defined.")
    def test_condor_python_exec(self):
        problem = PythonExecProblem()

        individuals = [Individual([1, 2])]
        algorithm = DummyAlgorithm(problem)
        algorithm.evaluator.evaluate(individuals)

        self.assertAlmostEqual(5, individuals[0].costs[0])

    @unittest.skipIf(config["condor_host"] is None, "Condor is not defined.")
    def test_condor_python_exec_full_load(self):
        problem = PythonExecProblem()

        n = 50
        individuals = []
        for i in range(n):
            individuals.append(Individual([random.randrange(1, 100), random.randrange(1, 100)]))

        algorithm = DummyAlgorithm(problem)
        algorithm.options['max_processes'] = 70
        algorithm.evaluator.evaluate(individuals)

        for i in range(n):
            # print(individuals[i])
            self.assertEqual(int(individuals[i].costs[0]), individuals[i].vector[0]**2 + individuals[i].vector[1]**2)

    @unittest.skipIf(config["condor_host"] is None, "Condor is not defined.")
    def test_condor_python_input(self):
        problem = PythonInputProblem()

        individuals = [Individual([1, 2])]
        algorithm = DummyAlgorithm(problem)
        algorithm.evaluator.evaluate(individuals)

        self.assertAlmostEqual(5, individuals[0].costs[0])


    @unittest.skipIf(config["condor_host"] is None, "Condor is not defined.")
    def test_condor_python_exec_nsgaii(self):
        problem = PythonExecProblemNSGAII()

        algorithm = NSGAII(problem)
        algorithm.options['max_population_number'] = 5
        algorithm.options['max_population_size'] = 3
        algorithm.options['max_processes'] = 3
        algorithm.run()

        populations = problem.populations()

        self.assertEqual(len(populations), algorithm.options['max_population_number'] + 1)
        for individuals in populations.values():
            self.assertEqual(len(individuals), algorithm.options['max_population_size'])


if __name__ == '__main__':
    main()



from ..operators import SimpleMutator, SimulatedBinaryCrossover, SimpleCrossover, \
    TournamentSelector, ParetoDominance, nondominated_truncate, crowding_distance, PmMutator, EpsilonDominance, \
    UniformMutator, NonUniformMutation, FireflyStep
from ..individual import Individual
from ..benchmark_pareto import BiObjectiveTestProblem
from ..problem import Problem
from math import inf
import unittest


class TestCrossover(unittest.TestCase):
    """ Tests crossover."""

    def setUp(self):
        self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [0, 5]},
                           {'name': 'x_2', 'initial_value': 1.5, 'bounds': [0, 3]}]

        self.signs = [1, 1]
        self.signs2 = [1, 1, 1]

        self.i1 = Individual([1, 2, 2])
        self.i2 = Individual([3, 2, 1])

        self.features = {}
        self.features['dominate'] = set()
        self.features['crowding_distance'] = 0
        self.features['domination_counter'] = 0
        self.features['front_number'] = 0
        self.features['precision'] = 7
        self.features['feasible'] = 0.0

    def test_simple_mutation(self):
        sm = SimpleMutator(self.parameters, 0.7)
        individual = sm.mutate(self.i1)
        self.assertTrue(
            self.parameters[0]['bounds'][0] <= individual.vector[0] <= self.parameters[0]['bounds'][1] and
            self.parameters[1]['bounds'][0] <= individual.vector[1] <= self.parameters[1]['bounds'][1])

    def test_sbx(self):
        sbx = SimulatedBinaryCrossover(self.parameters, 1.0)
        offsprings = sbx.cross(self.i1, self.i2)
        self.assertEqual(len(offsprings), 2)

    def test_simple_crossover(self):
        sbx = SimpleCrossover(self.parameters, 0.9)
        offsprings = sbx.cross(self.i1, self.i2)
        self.assertEqual(len(offsprings), 2)

    def test_dominates(self):
        dominance = ParetoDominance()
        i1 = Individual([0, 0])
        i1.costs = [1, 1]
        i1.features = self.features.copy()
        i1.calc_signed_costs(self.signs)

        i2 = Individual([2, 2])
        i2.costs = [2, 2]
        i2.features = self.features.copy()
        i2.calc_signed_costs(self.signs)

        # test weak-strong dominance
        i3 = Individual([0, 0, 0])
        i3.costs = [1, 0, 2]
        i3.features = self.features.copy()
        i3.calc_signed_costs(self.signs2)

        i4 = Individual([1, 1, 1])
        i4.costs = [1, 1, 1]
        i4.features = self.features.copy()
        i4.calc_signed_costs(self.signs2)

        result = dominance.compare(i1.costs_signed, i2.costs_signed)
        self.assertEqual(result, 1)

        result = dominance.compare(i2.costs_signed, i1.costs_signed)
        self.assertEqual(result, 2)

        result = dominance.compare(i3.costs_signed, i4.costs_signed)
        self.assertEqual(result, 0)

    def test_pareto(self):
        individuals = []
        n = 10
        for i in range(n):
            for j in range(i + 2):
                individual = Individual([j, i + 1 - j])
                individual.costs = [j, i + 1 - j]
                individual.features = self.features.copy()
                individuals.append(individual)

        for individual in individuals:
            individual.calc_signed_costs(self.signs2)

        selector = TournamentSelector(self.parameters)
        selector.fast_nondominated_sorting(individuals)

        for individual in individuals:
            self.assertEqual(individual.costs_signed[0] + individual.costs_signed[1],
                             individual.features['front_number'])


class TestFastNonDominatedSorting(unittest.TestCase):

    def setUp(self):
        test2d = BiObjectiveTestProblem()
        self.selector = TournamentSelector(test2d.parameters)

    def test_should_constructor_create_a_valid_object(self):
        self.assertIsNotNone(self.selector)

    def test_should_compute_crowding_distance_if_the_population_contains_one_solution(self):
        x = Individual([0, 0])
        x.costs = [2, 3]
        population = [x]

        self.selector.fast_nondominated_sorting(population)
        self.assertAlmostEqual(inf, population[0].features['crowding_distance'])

    def test_should_compute_crowding_distance_if_the_population_contains_two_individuals(self):
        x = Individual([0, 0])
        x.costs = [2, 3]
        x.costs_signed = [2, 3, 0]
        y = Individual([1, 1, 0])  # last index means that the solution is computed correctly
        y.costs = [1, 1]
        y.costs_signed = [1, 1, 0]

        population = [x, y]

        self.selector.fast_nondominated_sorting(population)
        self.assertAlmostEqual(inf, population[0].features['crowding_distance'])
        self.assertAlmostEqual(inf, population[1].features['crowding_distance'])

    def test_should_compute_ranking_work_properly_case1(self):
        """ The list contains two solutions and y is dominated by x."""
        x = Individual([2, 2])
        x.costs = [2, 3]
        x.costs_signed = [2, 3, 0]
        y = Individual([2, 2, 0])  # last index means that the solution is computed correctly
        y.costs = [3, 6]
        y.costs_signed = [3, 6, 0]

        population = [x, y]
        self.selector.fast_nondominated_sorting(population)
        nondominated_truncate(population, 2)
        self.assertEqual(population[0].features['front_number'], 1)
        self.assertEqual(population[1].features['front_number'], 2)

    def test_should_sort_the_population_with_three_dominated_solutions_return_three_subfronts(self):
        x = Individual([2, 2])
        x.costs = [2, 3]
        x.costs_signed = [2, 3, 0]

        y = Individual([2, 2])  # last index means that the solution is computed correctly
        y.costs = [3, 6]
        y.costs_signed = [3, 6, 0]

        z = Individual([2, 2, 0])  # last index means that the solution is computed correctly
        z.costs = [4, 8]
        z.costs_signed = [4, 8, 0]

        population = [x, y, z]
        self.selector.fast_nondominated_sorting(population)
        nondominated_truncate(population, 3)
        self.assertEqual(population[0].features['front_number'], 1)
        self.assertEqual(population[1].features['front_number'], 2)
        self.assertEqual(population[2].features['front_number'], 3)

        self.assertAlmostEqual(inf, population[0].features['crowding_distance'])
        self.assertAlmostEqual(inf, population[1].features['crowding_distance'])
        self.assertAlmostEqual(inf, population[2].features['crowding_distance'])

    def test_should_ranking_of_a_population_with_five_solutions_work_properly(self):
        x = Individual([1.0, 0.0])
        y = Individual([0.5, 0.5])
        z = Individual([0, 1.0])
        v = Individual([0.6, 0.6])
        w = Individual([0.7, 0.5])

        x.costs_signed = [1.0, 0.0, 0.0]  # third value: 0 means its calculated
        y.costs_signed = [0.5, 0.5, 0.0]
        z.costs_signed = [0.0, 1.0, 0.0]

        v.costs_signed = [0.6, 0.6, 0.0]
        w.costs_signed = [0.7, 0.5, 0.0]

        population = [x, y, z, v, w]

        self.selector.fast_nondominated_sorting(population)
        population = nondominated_truncate(population, 5)

        # [z, x, y]
        self.assertEqual(population[0].features['front_number'], 1)
        self.assertEqual(population[1].features['front_number'], 1)
        self.assertEqual(population[2].features['front_number'], 1)

        self.assertEqual(population[3].features['front_number'], 2)
        self.assertEqual(population[4].features['front_number'], 2)

        self.assertAlmostEqual(inf, population[0].features['crowding_distance'])
        self.assertAlmostEqual(inf, population[1].features['crowding_distance'])
        self.assertAlmostEqual(2.0, population[2].features['crowding_distance'])

        self.assertAlmostEqual(inf, population[3].features['crowding_distance'])
        self.assertAlmostEqual(inf, population[4].features['crowding_distance'])

    def test_should_the_crowding_distance_of_four_solutions_correctly_assigned(self):
        x = Individual([0, 1])
        y = Individual([1, 0])
        z = Individual([0.5, 0.5])
        v = Individual([0.75, 0.75])

        x.costs_signed = [0.0, 1.0, 0.0]  # third value: 0 means its calculated
        y.costs_signed = [1.0, 0.0, 0.0]
        z.costs_signed = [0.5, 0.5, 0.0]
        v.costs_signed = [0.75, 0.75, 0.0]

        population = [x, y, v, z]  # the expected results after sorting is : [x, y, z, v]

        for p in population:
            p.features['front_number'] = 1

        crowding_distance(population)
        population = nondominated_truncate(population, 4)

        self.assertAlmostEqual(inf, population[0].features['crowding_distance'])
        self.assertAlmostEqual(inf, population[1].features['crowding_distance'])
        self.assertAlmostEqual(1.5, population[2].features['crowding_distance'])
        self.assertAlmostEqual(1.0, population[3].features['crowding_distance'])


class SBXCrossoverTestCases(unittest.TestCase):

    def test_should_constructor_assign_the_correct_probability_value(self):
        crossover_probability = 0.8
        distribution_index = 15
        test2d = BiObjectiveTestProblem()
        SBXCrossover = SimulatedBinaryCrossover(test2d.parameters, crossover_probability, distribution_index)

        self.assertEqual(crossover_probability, SBXCrossover.probability)
        self.assertEqual(distribution_index, SBXCrossover.distribution_index)

    def test_should_constructor_raise_an_exception_if_the_probability_is_greater_than_one(self):
        with self.assertRaises(Exception):
            test2d = BiObjectiveTestProblem()
            SimulatedBinaryCrossover(test2d.parameters, 1.5, 20)

        with self.assertRaises(Exception):
            test2d = BiObjectiveTestProblem()
            SimulatedBinaryCrossover(test2d.parameters, 0.5, -1)


class DominanceComparator(unittest.TestCase):

    def test_should_dominance_comparator_return_zero_if_the_two_solutions_have_one_objective_with_the_same_value(self):
        dominance = ParetoDominance()
        x = Individual([2, 2])
        x.costs = [2.]
        x.costs_signed = [2., 0]

        y = Individual([2, 2])  # last index means that the solution is computed correctly
        y.costs = [2.]
        y.costs_signed = [2., 0]

        result = dominance.compare(x.costs_signed, y.costs_signed)

        self.assertEqual(0, result)

    def test_should_dominance_comparator_return_two_if_the_two_solutions_have_one_objective_and_the_second_one_is_lower(
            self):
        dominance = ParetoDominance()
        x = Individual([2, 2])
        x.costs = [2.]
        x.costs_signed = [2., 0]

        y = Individual([2, 2])  # last index means that the solution is computed correctly
        y.costs = [1.]
        y.costs_signed = [1., 0]

        result = dominance.compare(x.costs_signed, y.costs_signed)

        self.assertEqual(2, result)

    def test_should_dominance_comparator_return_one_if_the_two_solutions_have_one_objective_and_the_first_one_is_lower(
            self):
        dominance = ParetoDominance()
        x = Individual([2, 2])
        x.costs = [1.]
        x.costs_signed = [1., 0]

        y = Individual([2, 2])  # last index means that the solution is computed correctly
        y.costs = [2.]
        y.costs_signed = [2., 0]

        result = dominance.compare(x.costs_signed, y.costs_signed)

        self.assertEqual(1, result)

    def test_should_dominance_comparator_work_properly_case_a(self):
        """ Case A: solution1 has objectives [-1.0, 5.0, 9.0] and solution2 has [2.0, 6.0, 15.0]
        """

        dominance = ParetoDominance()
        x = Individual([2, 2])
        x.costs = [-1.0, 5.0, 9.0]
        x.costs_signed = [-1.0, 5.0, 9.0, 0]

        y = Individual([2, 2])  # last index means that the solution is computed correctly
        y.costs = [2.0, 6.0, 15.0]
        y.costs_signed = [2.0, 6.0, 15.0, 0]

        result = dominance.compare(x.costs_signed, y.costs_signed)

        self.assertEqual(1, result)

    def test_should_dominance_comparator_work_properly_case_b(self):
        """ Case B: solution1 has objectives [-1.0, 5.0, 9.0] and solution2 has [-1.0, 5.0, 10.0]
        """

        dominance = ParetoDominance()
        x = Individual([2, 2])
        x.costs = [-1.0, 5.0, 9.0]
        x.costs_signed = [-1.0, 5.0, 9.0, 0]

        y = Individual([2, 2])  # last index means that the solution is computed correctly
        y.costs = [-1.0, 5.0, 10.0]
        y.costs_signed = [-1.0, 5.0, 10.0, 0]

        result = dominance.compare(x.costs_signed, y.costs_signed)

        self.assertEqual(1, result)

    def test_should_dominance_comparator_work_properly_case_c(self):
        """ Case C: solution1 has objectives [-1.0, 5.0, 9.0] and solution2 has [-2.0, 5.0, 9.0]
        """
        dominance = ParetoDominance()
        x = Individual([2, 2])
        x.costs = [-1.0, 5.0, 9.0]
        x.costs_signed = [-1.0, 5.0, 9.0, 0]

        y = Individual([2, 2])  # last index means that the solution is computed correctly
        y.costs = [-2.0, 5.0, 9.0]
        y.costs_signed = [-2.0, 5.0, 9.0, 0]

        result = dominance.compare(x.costs_signed, y.costs_signed)

        self.assertEqual(2, result)

    def test_should_dominance_comparator_work_properly_case_d(self):
        """ Case d: solution1 has objectives [-1.0, 5.0, 9.0] and solution2 has [-1.0, 5.0, 8.0]
        """
        dominance = ParetoDominance()
        x = Individual([2, 2])
        x.costs = [-1.0, 5.0, 9.0]
        x.costs_signed = [-1.0, 5.0, 9.0, 0]

        y = Individual([2, 2])  # last index means that the solution is computed correctly
        y.costs = [-1.0, 5.0, 8.0]
        y.costs_signed = [-1.0, 5.0, 8.0, 0]

        result = dominance.compare(x.costs_signed, y.costs_signed)

        self.assertEqual(2, result)

    def test_should_dominance_comparator_work_properly_with_constrains_case_1(self):
        """ Case 1: solution2 has a higher degree of constraint violation than solution 1
        """
        dominance = ParetoDominance()
        x = Individual([2, 2])
        x.costs = [-1.0, 6.0, 11.0]
        x.costs_signed = [-1.0, 6.0, 11.0, -0.1]

        y = Individual([2, 2])  # last index means that the solution is computed correctly
        y.costs = [-1.0, 5.0, 10.0]
        y.costs_signed = [-1.0, 5.0, 10.0, -0.3]

        result = dominance.compare(x.costs_signed, y.costs_signed)
        self.assertEqual(1, result)

    def test_should_dominance_comparator_work_properly_with_constrains_case_2(self):
        """ Case 2: solution1 has a higher degree of constraint violation than solution 1
        """
        dominance = ParetoDominance()
        x = Individual([2, 2])
        x.costs = [-1.0, 6.0, 9.0]
        x.costs_signed = [-1.0, 6.0, 9.0, -0.5]

        y = Individual([2, 2])  # last index means that the solution is computed correctly
        y.costs = [-1.0, 6.0, 10.0]
        y.costs_signed = [-1.0, 6.0, 10.0, -0.1]

        result = dominance.compare(x.costs_signed, y.costs_signed)
        self.assertEqual(2, result)


class EpsDominanceComparator(unittest.TestCase):

    def test_should_dominance_comparator_work_properly_case_d(self):
        """ Case 1: solution1 has objectives [-1.0, 5.0, 9.0] and solution2 has [-1.0, 5.0, 8.0]
        """

        # eps-dominance comparator should give back the same results as pareto-dominance if eps = 1e-5

        dominance = EpsilonDominance(epsilons=1e-2)
        x = Individual([2, 2])
        x.costs = [-1.0, 5.0, 9.0]
        x.costs_signed = [-1.0, 5.0, 9.0, 0]

        y = Individual([2, 2])  # last index means that the solution is computed correctly
        y.costs = [-1.0, 5.01, 8.0]
        y.costs_signed = [-1.0, 5.01, 8.0, 0]

        result = dominance.compare(x.costs_signed, y.costs_signed)

        self.assertEqual(2, result)

        # eps-dominance comparator should give different results if eps = 1

        dominance = EpsilonDominance(epsilons=1e-3)
        x = Individual([2, 2])
        x.costs = [-1.0, 5.0, 9.0]
        x.costs_signed = [-1.0, 5.0, 9.0, 0]

        y = Individual([2, 2])  # last index means that the solution is computed correctly
        y.costs = [-1.0, 5.01, 8.0]
        y.costs_signed = [-1.0, 5.01, 8.0, 0]

        result = dominance.compare(x.costs_signed, y.costs_signed)

        self.assertEqual(0, result)

    def test_should_dominance_comparator_work_properly_with_constrains_case_1(self):
        """ Case 1: solution2 has a higher degree of constraint violation than solution 1
        """
        dominance = EpsilonDominance(epsilons=1e-5)
        x = Individual([2, 2])
        x.costs = [-1.0, 6.0, 11.0]
        x.costs_signed = [-1.0, 6.0, 11.0, -0.1]

        y = Individual([2, 2])  # last index means that the solution is computed correctly
        y.costs = [-1.0, 5.0, 10.0]
        y.costs_signed = [-1.0, 5.0, 10.0, -0.3]

        result = dominance.compare(x.costs_signed, y.costs_signed)
        self.assertEqual(1, result)

    def test_should_dominance_comparator_work_properly_with_constrains_case_2(self):
        """ Case 2: solution1 has a higher degree of constraint violation than solution 1
        """
        dominance = EpsilonDominance(epsilons=1e-5)
        x = Individual([2, 2])
        x.costs = [-1.0, 6.0, 9.0]
        x.costs_signed = [-1.0, 6.0, 9.0, -0.5]

        y = Individual([2, 2])  # last index means that the solution is computed correctly
        y.costs = [-1.0, 6.0, 10.0]
        y.costs_signed = [-1.0, 6.0, 10.0, -0.1]

        result = dominance.compare(x.costs_signed, y.costs_signed)
        self.assertEqual(2, result)

    def test_should_dominance_comparator_work_properly_select_closer_to_boundary(self):
        """ Case 3: solution1 closer to the rounded value than solution 2
        """
        dominance = EpsilonDominance(epsilons=0.1)
        x = Individual([2, 2])
        x.costs = [-1.0, 6.0, 9.05]
        x.costs_signed = [-1.0, 6.0, 9.05, 0.0]

        y = Individual([2, 2])  # last index means that the solution is computed correctly
        y.costs = [-1.0, 6.0, 9.06]
        y.costs_signed = [-1.0, 6.0, 9.06, 0.0]

        result = dominance.compare(x.costs_signed, y.costs_signed)
        self.assertEqual(1, result)


class TestTournamentSelector(unittest.TestCase):

    def test_should_execute_work_properly_case1(self):
        x = Individual([3, 2])
        x.costs = [2.0, 3.0]
        x.costs_signed = [2.0, 3.0, 0.0]

        y = Individual([3, 2])  # last index means that the solution is computed correctly
        y.costs = [1.0, 4.0]
        y.costs_signed = [1.0, 4.0, 0.0]


class TestProblem(Problem):

    def set(self):
        self.name = 'Test Problem'
        self.parameters = [{'name': '1', 'bounds': [2., 10.]},
                           {'name': '2', 'bounds': [2., 20.]},
                           ]

        self.costs = [{'name': 'f_1', 'criteria': 'minimize'}]

    def evaluate(self, individual):
        return [1, 1]


class TestUniformMutator(unittest.TestCase):

    def setUp(self):
        problem = TestProblem()
        self.um = UniformMutator(problem.parameters, 0)

    def test_should_the_solution_remain_unchanged_if_the_probability_is_zero(self):
        x = Individual([3, 2])
        x.costs = [2.0, 3.0]
        x.costs_signed = [2.0, 3.0, 0.0]

        y = self.um.mutate(x)
        self.assertEqual(y.vector, [3, 2])

    def test_should_change_if_the_probability_is_one(self):
        x = Individual([3, 2])
        x.costs = [2.0, 3.0]
        x.costs_signed = [2.0, 3.0, 0.0]

        self.um.probability = 1.0
        y = self.um.mutate(x)
        self.assertNotEqual(y.vector, [3, 2])


class TestUniformMutator(unittest.TestCase):

    def setUp(self):
        problem = TestProblem()
        self.um = PmMutator(problem.parameters, 0)

    def test_should_the_solution_remain_unchanged_if_the_probability_is_zero(self):
        x = Individual([3, 2])
        x.costs = [2.0, 3.0]
        x.costs_signed = [2.0, 3.0, 0.0]

        y = self.um.mutate(x)
        self.assertEqual(y.vector, [3, 2])

    def test_should_change_if_the_probability_is_one(self):
        x = Individual([3, 2])
        x.costs = [2.0, 3.0]
        x.costs_signed = [2.0, 3.0, 0.0]

        self.um.probability = 1.0
        y = self.um.mutate(x)
        self.assertNotEqual(y.vector, [3, 2])


class TestNonUniformMutator(unittest.TestCase):

    def setUp(self):
        problem = TestProblem()
        self.um = NonUniformMutation(problem.parameters, 0, 100, 1)

    def test_should_the_solution_remain_unchanged_if_the_probability_is_zero(self):
        x = Individual([3, 2])
        x.costs = [2.0, 3.0]
        x.costs_signed = [2.0, 3.0, 0.0]

        y = self.um.mutate(x)
        self.assertEqual(y.vector, [3, 2])

    def test_should_change_if_the_probability_is_one(self):
        x = Individual([3, 2])
        x.costs = [2.0, 3.0]
        x.costs_signed = [2.0, 3.0, 0.0]

        self.um.probability = 1.0
        y = self.um.mutate(x, 2)
        self.assertNotEqual(y.vector, [3, 2])


class TestFireflystep(unittest.TestCase):

    def setUp(self):
        problem = TestProblem()
        self.crossover = FireflyStep(problem.parameters, 0, 1.0, 0, 1)

    def test_should_not_change_the_vector_by_the_given_values_if_not_dominates(self):
        x = Individual([4, 3])
        x.costs = [2.0, 3.0]
        x.costs_signed = [2.0, 3.0, 0.0]

        y = Individual([3, 2])
        y.costs = [1.0, 2.0]
        y.costs_signed = [1.0, 1.0, 0.0]

        self.crossover.attraction_step(y, x, 1)
        self.assertEqual([3.0, 2.0], y.vector)

    def test_should_change_the_vector_by_the_given_values(self):
        x = Individual([4, 3])
        x.costs = [2.0, 3.0]
        x.costs_signed = [2.0, 3.0, 0.0]

        y = Individual([3, 2])
        y.costs = [1.0, 2.0]
        y.costs_signed = [1.0, 1.0, 0.0]

        self.crossover.attraction_step(x, y, 1)
        self.assertEqual([3.0, 2.0], x.vector)


if __name__ == '__main__':
    unittest.main()



from .utils import VectorAndNumbers
from .algorithm import Algorithm
from .individual import Individual
from .job import Job
from .problem import Problem
from .operators import CustomGenerator

from SALib.sample.saltelli import sample as sobol_sample
from SALib.sample.morris import sample as morris_sample
from SALib.sample.fast_sampler import sample as fast_sample
from SALib.sample.latin import sample as latin_sample
from SALib.sample.ff import sample as ff_sample

from SALib.analyze import sobol
from SALib.analyze import ff
from SALib.analyze import morris
from SALib.analyze import fast
from SALib.analyze import rbd_fast
from SALib.analyze import delta

import time
import numpy as np

_method = ['rbd_fast', 'fast', 'morris', 'sobol', 'delta', 'ff']


class SALibAlgorithm(Algorithm):
    """
    SALib Analysis
    """

    def __init__(self, problem: Problem, name='SALibAlgorithm'):
        super().__init__(problem, name)

        self.sa_problem = {}
        self.samples_x = []
        self.samples_y = []

        self.job = Job(self.problem)

        self.options.declare(name='method', default='sobol', values=_method,
                             desc='Method')
        self.options.declare(name='print_to_console', default=False,
                             desc='Print to console')
        self.options.declare(name='samples', default=10, lower=1,
                             desc='Samples')

    def run(self):
        t_s = time.time()

        # set SALib problem
        names = []
        bounds = []
        for parameter in self.problem.parameters:
            names.append(parameter['name'])
            bounds.append(parameter['bounds'])

        self.sa_problem = {'num_vars': len(self.problem.parameters),
                           'names': names,
                           'bounds': bounds}

        # generate samples
        if self.options["method"] == "rbd_fast":
            self.samples_x = latin_sample(self.sa_problem, self.options["samples"])
        elif self.options["method"] == "fast":
            self.samples_x = fast_sample(self.sa_problem, self.options["samples"])
        elif self.options["method"] == "morris":
            self.samples_x = morris_sample(self.sa_problem, self.options["samples"], num_levels=4)
        elif self.options["method"] == "sobol":
            self.samples_x = sobol_sample(self.sa_problem, self.options["samples"])
        elif self.options["method"] == "delta":
            self.samples_x = latin_sample(self.sa_problem, self.options["samples"])
        elif self.options["method"] == "ff":
            self.samples_x = ff_sample(self.sa_problem, self.options["samples"])

        individuals = []
        for vector in self.samples_x:
            individuals.append(Individual(vector))

        # append to problem
        for individual in individuals:
            self.problem.individuals.append(individual)

        # evaluate individuals
        self.evaluate(individuals)

        for individual in individuals:
            self.samples_y.append(individual.costs[0]) # TODO: fix index [0]
        self.samples_y = np.array(self.samples_y)

        t = time.time() - t_s
        self.problem.logger.info("Sensitivity: elapsed time: {} s".format(t))

        # sync changed individual informations
        self.problem.data_store.sync_all()


    def analyze(self):
        if self.options["method"] == "rbd_fast":
            return self.analyze_rbd_fast()
        elif self.options["method"] == "fast":
            return self.analyze_sobol()
        elif self.options["method"] == "morris":
            return self.analyze_morris()
        elif self.options["method"] == "sobol":
            return self.analyze_sobol()
        elif self.options["method"] == "delta":
            return self.analyze_delta()
        elif self.options["method"] == "ff":
            return self.analyze_ff()

    def analyze_rbd_fast(self):
        #  RBD-FAST - Random Balance Designs Fourier Amplitude Sensitivity Test
        return rbd_fast.analyze(self.sa_problem, self.samples_x, self.samples_y, print_to_console=self.options["print_to_console"])

    def analyze_fast(self):
        # FAST - Fourier Amplitude Sensitivity Test
        return fast.analyze(self.sa_problem, self.samples_y, print_to_console=self.options["print_to_console"])

    def analyze_morris(self):
        # Method of Morris
        return morris.analyze(self.sa_problem, self.samples_x, self.samples_y, conf_level=0.95, num_levels=4, print_to_console=self.options["print_to_console"])

    def analyze_sobol(self):
        # Sobol Sensitivity Analysis
        return sobol.analyze(self.sa_problem, self.samples_y, print_to_console=self.options["print_to_console"])

    def analyze_delta(self):
        # Delta Moment-Independent Measure
        return delta.analyze(self.sa_problem, self.samples_x, self.samples_y, print_to_console=self.options["print_to_console"])

    def analyze_ff(self):
        # Fractional Factorial
        return ff.analyze(self.sa_problem, self.samples_x, self.samples_y, second_order=True, print_to_console=self.options["print_to_console"])


class Sensitivity(Algorithm):
    def __init__(self, problem, parameters, name='Sensitivity analysis'):
        self.parameters = parameters
        super().__init__(problem, name)

        self.options.declare(name='max_population_size', default=100, lower=1,
                             desc='Maximal number of individuals in population')

    def run(self):
        parameters = []
        for parameter in self.problem.parameters:
            parameters.append(float(parameter['initial_value']))

        for parameter_name in self.parameters:
            parameter_values = []

            index = 0
            selected_parameter = None
            for parameter in self.parameters:
                if parameter['name'] == parameter_name['name']:
                    selected_parameter = parameter
                    break
                index += 1

            individuals = []
            for i in range(self.options['max_population_size']):
                value = VectorAndNumbers.gen_number(selected_parameter['bounds'], selected_parameter['precision'], 'normal')
                parameters[index] = value
                parameter_values.append(value)
                individual = Individual(parameters.copy())
                individuals.append(individual)

            self.evaluate(individuals)
            # costs = []
            # # TODO: Make also for multi-objective
            # for individual in individuals:
            #     costs.append(individual.costs)

        # append individuals
        for individual in individuals:
            self.problem.individuals.append(individual)

        # sync changed individual informations
        self.problem.data_store.sync_all()




import unittest
from ..operators import CustomGenerator, RandomGenerator, FullFactorGenerator, FullFactorLevelsGenerator, PlackettBurmanGenerator, \
    BoxBehnkenGenerator, LHSGenerator, GSDGenerator, HaltonGenerator, UniformGenerator


class TestDOE(unittest.TestCase):

    def setUp(self):
        self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [-2.5, 5]},
                           {'name': 'x_2', 'initial_value': 1.5, 'bounds': [1, 3.4]},
                           {'name': 'x_3', 'initial_value': 3.5, 'bounds': [6, 10]}]

    def test_random_generator(self):
        gen = RandomGenerator(parameters=self.parameters)
        gen.init(3)
        individuals = gen.generate()
        # values
        self.assertTrue(self.parameters[0]['bounds'][0] <= individuals[0].vector[0] <= self.parameters[0]['bounds'][1] and
                        self.parameters[1]['bounds'][0] <= individuals[0].vector[1] <= self.parameters[1]['bounds'][1] and
                        self.parameters[2]['bounds'][0] <= individuals[0].vector[2] <= self.parameters[2]['bounds'][1] and
                        self.parameters[0]['bounds'][0] <= individuals[1].vector[0] <= self.parameters[0]['bounds'][1] and
                        self.parameters[1]['bounds'][0] <= individuals[1].vector[1] <= self.parameters[1]['bounds'][1] and
                        self.parameters[2]['bounds'][0] <= individuals[1].vector[2] <= self.parameters[2]['bounds'][1] and
                        self.parameters[0]['bounds'][0] <= individuals[2].vector[0] <= self.parameters[0]['bounds'][1] and
                        self.parameters[1]['bounds'][0] <= individuals[2].vector[1] <= self.parameters[1]['bounds'][1] and
                        self.parameters[2]['bounds'][0] <= individuals[2].vector[2] <= self.parameters[2]['bounds'][1])

    def test_custom_generator(self):
        gen = CustomGenerator()
        gen.init([[3, 2, 6], [-1, 3, 8]])
        individuals = gen.generate()
        # values
        self.assertEqual(individuals[0].vector[0], 3)
        self.assertEqual(individuals[0].vector[1], 2)
        self.assertEqual(individuals[0].vector[2], 6)
        self.assertEqual(individuals[1].vector[0], -1)
        self.assertEqual(individuals[1].vector[1], 3)
        self.assertEqual(individuals[1].vector[2], 8)

    def test_uniform_generator(self):
        gen = UniformGenerator(parameters=self.parameters)
        gen.init(4)
        individuals = gen.generate()

        # values
        self.assertEqual(len(individuals), 4**len(self.parameters))
        self.assertEqual(individuals[0].vector[2], 6.0)
        self.assertEqual(individuals[1].vector[0], -2.5)
        self.assertEqual(individuals[12].vector[1], 3.4)
        self.assertEqual(individuals[28].vector[2], 6.0)
        self.assertEqual(individuals[35].vector[0], 2.5)
        self.assertEqual(individuals[63].vector[2], 10.0)

    # Factorial Designs
    def test_full_fact_generator(self):
        # General Full-Factorial
        gen = FullFactorGenerator(parameters=self.parameters)
        gen.init(center=False)
        individuals = gen.generate()

        # size
        self.assertEqual(len(individuals), 8)
        # values
        self.assertEqual(individuals[0].vector[0], -2.5)
        self.assertEqual(individuals[1].vector[0], 5.0)
        self.assertEqual(individuals[2].vector[1], 3.4)
        self.assertEqual(individuals[3].vector[1], 3.4)
        self.assertEqual(individuals[7].vector[2], 10.0)

        gen = FullFactorGenerator(parameters=self.parameters)
        gen.init(center=True)
        individuals = gen.generate()

        self.assertEqual(individuals[3].vector[0], -2.5)
        self.assertEqual(individuals[-1].vector[1], 3.4)

    def test_plackett_burman_generator(self):
        # Plackett-Burman
        gen = PlackettBurmanGenerator(parameters=self.parameters)
        individuals = gen.generate()

        # size
        self.assertEqual(len(individuals), 4)
        # values
        self.assertEqual(individuals[0].vector[0], -2.5)
        self.assertEqual(individuals[1].vector[0], 5.0)
        self.assertEqual(individuals[2].vector[2], 6.0)
        self.assertEqual(individuals[3].vector[1], 3.4)

    # Response - Surface Designs
    def test_box_behnken_generator(self):
        # Box-Behnken
        gen = BoxBehnkenGenerator(parameters=self.parameters)
        individuals = gen.generate()

        # size
        self.assertEqual(len(individuals), 13)
        # values
        self.assertEqual(individuals[2].vector[2], 8.0)
        self.assertEqual(individuals[3].vector[1], 3.4)
        self.assertEqual(individuals[5].vector[0], 5.0)
        self.assertEqual(individuals[12].vector[1], 2.2)

    # Randomized Designs
    def test_lhs_generation(self):
        # Latin - Hypercube
        gen = LHSGenerator(parameters=self.parameters)
        gen.init(number=3)
        individuals = gen.generate()

        # size
        self.assertEqual(len(individuals), 3)
        # values
        self.assertTrue(self.parameters[0]['bounds'][0] <= individuals[0].vector[0] <= self.parameters[0]['bounds'][1] and
                        self.parameters[1]['bounds'][0] <= individuals[0].vector[1] <= self.parameters[1]['bounds'][1] and
                        self.parameters[2]['bounds'][0] <= individuals[0].vector[2] <= self.parameters[2]['bounds'][1] and
                        self.parameters[0]['bounds'][0] <= individuals[1].vector[0] <= self.parameters[0]['bounds'][1] and
                        self.parameters[1]['bounds'][0] <= individuals[1].vector[1] <= self.parameters[1]['bounds'][1] and
                        self.parameters[2]['bounds'][0] <= individuals[1].vector[2] <= self.parameters[2]['bounds'][1] and
                        self.parameters[0]['bounds'][0] <= individuals[2].vector[0] <= self.parameters[0]['bounds'][1] and
                        self.parameters[1]['bounds'][0] <= individuals[2].vector[1] <= self.parameters[1]['bounds'][1] and
                        self.parameters[2]['bounds'][0] <= individuals[2].vector[2] <= self.parameters[2]['bounds'][1])

    # Randomized Designs
    def test_halton_generation(self):
        # Halton
        gen = HaltonGenerator(parameters=self.parameters)
        gen.init(number=3)
        individuals = gen.generate()

        # size
        self.assertEqual(len(individuals), 3)
        # values
        self.assertTrue(self.parameters[0]['bounds'][0] <= individuals[0].vector[0] <= self.parameters[0]['bounds'][1] and
                        self.parameters[1]['bounds'][0] <= individuals[0].vector[1] <= self.parameters[1]['bounds'][1] and
                        self.parameters[2]['bounds'][0] <= individuals[0].vector[2] <= self.parameters[2]['bounds'][1] and
                        self.parameters[0]['bounds'][0] <= individuals[1].vector[0] <= self.parameters[0]['bounds'][1] and
                        self.parameters[1]['bounds'][0] <= individuals[1].vector[1] <= self.parameters[1]['bounds'][1] and
                        self.parameters[2]['bounds'][0] <= individuals[1].vector[2] <= self.parameters[2]['bounds'][1] and
                        self.parameters[0]['bounds'][0] <= individuals[2].vector[0] <= self.parameters[0]['bounds'][1] and
                        self.parameters[1]['bounds'][0] <= individuals[2].vector[1] <= self.parameters[1]['bounds'][1] and
                        self.parameters[2]['bounds'][0] <= individuals[2].vector[2] <= self.parameters[2]['bounds'][1])
        # values
        self.assertEqual(individuals[0].vector[2], 6.8)
        self.assertEqual(individuals[2].vector[0], 3.125)

    # Factorial Designs - levels
    def test_full_fact_levels_generator(self):
        # GSD
        self.parameters = [{'name': 'U_1'},
                           {'name': 'U_2'},
                           {'name': 'U_3'},
                           {'name': 'PH_1'},
                           {'name': 'PH_2'},
                           {'name': 'PH_3'}]
        gen = FullFactorLevelsGenerator(parameters=self.parameters)
        amp = [0, 0.25, 0.5, 0.75, 1]
        phase = [-90.000, -67.500, -45.000, -22.500, 0.000, 22.500, 45.000, 67.500, 90.000]
        gen.init([amp, amp, amp, phase, phase, phase])
        individuals = gen.generate()

        # size
        self.assertEqual(len(individuals), 91125)
        # values
        self.assertEqual(individuals[152].vector[1], 0)
        self.assertEqual(individuals[8596].vector[0], 0.25)
        self.assertEqual(individuals[70153].vector[3], -22.5)
        self.assertEqual(individuals[90454].vector[5], 90.0)

    # Generalized Subset Design (GSD) (C) 2018 - Rickard Sjoegren
    def test_gsd_generator(self):
        # GSD
        # self.parameters = [{'name': 'U_1'}, {'name': 'U_2'}, {'name': 'U_3'}, {'name': 'PH_1'}, {'name': 'PH_2'}, {'name': 'PH_3'}]
        self.parameters = [{'name': 'U_1'}, {'name': 'U_2'}]
        gen = GSDGenerator(parameters=self.parameters)
        gen.init([[1, 3, 2], [6, 8, 4]], reduction=2)
        # amp = [0, 0.25, 0.5, 0.75, 1]
        # phase = [-90.000, -67.500, -45.000, -22.500, 0.000, 22.500, 45.000, 67.500, 90.000]
        # gen.init([amp, amp, amp, phase, phase, phase], reduction=10)
        individuals = gen.generate()

        # size
        self.assertEqual(len(individuals), 5)
        # values
        self.assertEqual(individuals[0].vector[1], 6)
        self.assertEqual(individuals[2].vector[0], 2)
        self.assertEqual(individuals[4].vector[0], 3)


if __name__ == '__main__':
    unittest.main()



import csv
from .quality_indicator import gd, epsilon_add
from matplotlib import pylab, rc, axis


class Results:
    """ Class Results offers tools for interpreting calculated data. """

    def __init__(self, problem):
        self.problem = problem
        self.use_latex = True
        self.label_size = 16
        self.tick_size = 14
        self.title_size = 18
        self.legend = True
        self.width = 8
        self.hight = 5

        if self.use_latex:
            # allows using Latex
            rc('font', **{'family': ' DejaVu Sans', 'sans-serif': ['Helvetica']})
            rc('text', usetex=True)

        # sets the legend font size
        params = {'legend.fontsize': 14,
                  'legend.handlelength': 2}

        pylab.rcParams.update(params)

    def parameter_names(self):
        parameter_names = []
        for parameter in self.problem.parameters:
            parameter_names.append(parameter["name"])
        return parameter_names

    def parameter_number(self):
        return len(self.problem.parameters)

    def goal_number(self):
        return len(self.problem.costs)

    def goal_names(self):
        cost_names = []
        for cost in self.problem.costs:
            cost_names.append(cost["name"])
        return cost_names

    def parameter_index(self, name):
        try:
            return next((index for (index, d) in enumerate(self.problem.parameters) if d["name"] == name), None)
        except ValueError as e:
            raise ValueError('There is not a parameter with given name: {}'.format(name))

    def goal_index(self, name):
        try:
            return next((index for (index, d) in enumerate(self.problem.costs) if d["name"] == name), None)
            # return list(self.problem.costs).index(name)
        except ValueError as e:
            raise ValueError('There is not a cost function with given name: {}'.format(name))

    def parameters(self):
        out = []
        for individuals in self.problem.populations().values():
            for individual in individuals:
                out.append(individual.vector)
        return out

    def costs(self):
        out = []
        for individuals in self.problem.populations().values():
            for individual in individuals:
                out.append(individual.costs)
        return out

    def table(self, transpose=True):
        out = []
        for individuals in self.problem.populations().values():
            for individual in individuals:
                out.append(individual.vector + individual.costs)

        if transpose:
            return list(zip(*out))
        else:
            return out

    def population(self, population_id=-1):
        # Returns population with given index
        #if population_id >= len(self.problem.populations):
        #    raise ValueError('Index of population exceeds the number of populations.')
        if population_id == -1:
            individuals = self.problem.last_population()
        else:
            individuals = self.problem.population(population_id)
        return individuals

    def goal_on_index(self, name=None, population_id=-1):
        """ Returns a list of lists. The first list contains indexes of individuals in population,
            the other lists contains values of the goal function(s).
        """
        individuals = self.population(population_id)
        n = range(len(individuals))

        table = [list(n)]
        if name is None:
            for j in range(len(self.problem.costs)):
                table.append([])
                for i in n:
                    table[j + 1].append(individuals[i].costs[j])
        else:
            table.append([])
            index = self.goal_index(name)
            for i in n:
                table[1].append(individuals[i].costs[index])

        return table

    def parameter_on_index(self, name=None, population_id=-1):
        """
        Returns a list of lists. The first list contains indexes of individuals in population,
            the other lists contains values of the parameters(s).
        :param name: string representing the name of parameter if it is not given, all parameters are included
        :param population_id: index of the selected parameter
        :return:
        """

        individuals = self.population(population_id)
        n = range(len(individuals))
        table = [list(n)]
        if name is None:
            for j in range(len(self.problem.parameters)):
                table.append([])
                for i in n:
                    table[j + 1].append(individuals[i].vector[j])
        else:
            table.append([])
            index = self.parameter_index(name)
            for i in n:
                table[1].append(individuals[i].vector[index])

        return table

    def goal_on_parameter(self, parameter_name, goal_name, population_id=-1, sorted=False):
        """
        The method returns the dependance of selected goal function on particular parameter
        :param parameter_name: string specifying particular parameter
        :param goal_name: string specifying name of required goal function
        :param population_id: index of required population, if it is not given the last population is used
        :return: list of two lists, the first contains sorted parameter values, the second values of selected goal
        function
        """

        if population_id == -1:
            individuals = self.problem.last_population()
        else:
            individuals = self.population(population_id)

        parameter_index = self.parameter_index(parameter_name)
        goal_index = self.goal_index(goal_name)
        parameter_values = []
        goal_values = []

        for individual in individuals:
            parameter_values.append(individual.vector[parameter_index])
            goal_values.append(individual.costs[goal_index])

        if sorted is True:
            goal_values = self.sort_list(parameter_values, goal_values)
            parameter_values.sort()

        return [parameter_values, goal_values]

    def parameter_on_goal(self, goal_name, parameter_name, population_id=-1, sorted=False):
        parameter_values, goal_values = self.goal_on_parameter(parameter_name, goal_name, population_id, sorted=False)
        if sorted:
            parameter_values = self.sort_list(goal_values, parameter_values)
            goal_values.sort()

        return [goal_values, parameter_values]

    def parameter_on_parameter(self, parameter_1, parameter_2, population_id=-1, sorted=False):
        individuals = self.population(population_id)
        index_1 = self.parameter_index(parameter_1)
        index_2 = self.parameter_index(parameter_2)
        values_1 = []
        values_2 = []

        for individual in individuals:
            values_1.append(individual.vector[index_1])
            values_2.append(individual.vector[index_2])

        if sorted:
            values_2 = self.sort_list(values_1, values_2)
            values_1.sort()

        return [values_1, values_2]

    def export_to_csv(self, filename):
        """
        Writes out populations into a csv file, the file contains the costs and the parameters
        :return:
        """
        with open(filename, 'w', newline='') as f:
            writer = csv.writer(f)
            # names
            out = ["population_id"]
            for parameter in self.problem.parameters:
                out.append(parameter["name"])
            for cost in self.problem.costs:
                out.append(cost["name"])

            writer.writerows([out])

            # values
            population_id = 0
            for individuals in self.problem.populations().values():
                for individual in individuals:
                    out = [population_id]
                    for i in individual.costs:
                        out.append(i)
                    for j in individual.vector:
                        out.append(j)

                    writer.writerows([out])
                population_id += 1

    def pareto_front(self, population_id=None):
        """

        :return: pareto front
        """
        if population_id is not None:
            individuals = self.population(population_id)
        else:
            individuals = self.problem.individuals

        n = self.goal_number()
        pareto_front = []
        for i in range(n):
            pareto_front.append([])
            for individual in individuals:
                if individual.features['front_number'] == 1:
                    pareto_front[i].append(individual.costs[i])

        return pareto_front

    @staticmethod
    def sort_list(list_1, list_2):
        zipped_pairs = zip(list_1, list_2)
        sorted_list = [x for _, x in sorted(zipped_pairs)]
        return sorted_list

    # TODO: Generalize
    # TODO: add test
    def find_optimum(self, name=None):
        """
        Search the optimal value for the given (by name parameter) single objective .

        :param name: string representing name of the goal function
        :return:
        """

        # get the index of the required parameter
        index = 0  # default - one parameter
        if name:
            for i in range(len(self.problem.costs)):
                cost = self.problem.costs[i]
                if cost['name'] == name:
                    index = i
        # if len(self.problem.data_store.individuals) is not 0:
        #    min_l = [min(self.problem.data_store.individuals, key=lambda x: x.costs[index])]
        # else:
        #if self.problem.archive:
        #    min_l = [min(self.problem.archive, key=lambda x: x.costs[index])]
        #else:
        if len(self.problem.individuals) > 0:
            min_l = [min(self.problem.individuals, key=lambda x: x.costs[index])]
        # for population in self.problem.populations:
        opt = min(min_l, key=lambda x: x.costs[index])
        return opt

    # TODO: same function - David, could you check it and write test?
    def pareto_values(self, archive=None):
        """
        :return: a list of lists which contains the optimal values of the cost function:
                l_sol[[c11, c12, ... c1n], ... [cm1, cm2, ... cmn]]
        """

        individuals = self.problem.last_population()
        l_sol = []

        # the pareto values collected in the archive, if the algorithm uses this strategy
        if archive:
            for individual in archive:
                l_sol.append(individual.costs)
        else:
            if len(individuals) > 1:
                for individual in individuals:
                    l_sol.append(individual.costs)

        return l_sol

    # TODO: add test
    def performance_measure(self, reference: list, type='epsilon'):
        """
        This function compares the result with a given, reference pareto-set. By default it offers to make
        an epsilon measure, but other metrics can be selected.

        :param reference: list of tuples, which are similar to the data of the calculated pareto-front
        :param type:
        :return:
        """

        computed = self.pareto_values()  # [[], [], .. ]

        if type == 'epsilon':
            result = epsilon_add(reference, computed)
        if type == 'gd':
            result = gd(reference, computed)

        return result

    def set_figure(self):
        # We change the fontsize of minor ticks label
        fig, ax = pylab.subplots(figsize=(self.width, self.hight))
        ax.tick_params(axis='both', which='major', labelsize=self.tick_size)
        ax.tick_params(axis='both', which='minor', labelsize=self.tick_size)
        ax.grid()
        # ax.grid(color='grey', linestyle='-', linewidth=0.25, alpha=0.5)
        # removing top and right borders
        # ax.spines['top'].set_visible(False)
        # ax.spines['right'].set_visible(False)  # adds major gridlines
        return [fig, ax]

    def get_objectives_plot(self):

        fig, ax = self.set_figure()
        all_individuals = self.costs()
        for point in all_individuals:
            ax.plot(point[0], point[1], '.')

        pareto = self.pareto_front()
        ax.plot(pareto[0], pareto[1], 'ro', label='Last Pareto front')

        ax.set_xlabel(r'$F_1$', fontsize=self.label_size)
        ax.set_ylabel(r'$F_2$', fontsize=self.label_size)
        ax.legend()
        return [fig, ax]

    def objectives_plot(self):
        fig, ax = self.get_objectives_plot()
        fig.show()

    def get_pareto_plot(self, cost_x=0, cost_y=1):
        """
        A simple 2d - pareto plot from the variable 1 and 2 by default.
        :return:
        """
        fig, ax = self.set_figure()
        pvalues = self.pareto_values()

        x = [i[cost_x] for i in pvalues]
        y = [i[cost_y] for i in pvalues]

        ax.scatter(x, y, alpha=0.80)

        # adds a title and axes labels
        x_name = self.problem.costs[cost_x]['name']
        y_name = self.problem.costs[cost_y]['name']

        ax.set_title('Pareto values', fontsize=self.title_size)
        ax.set_xlabel('$' + x_name + '$', fontsize=self.label_size)
        ax.set_ylabel('$' + y_name + '$', fontsize=self.label_size)
        return [fig, ax]


    def pareto_plot(self, cost_x=0, cost_y=1):
        fig, ax = self.get_pareto_plot(cost_x, cost_y)
        fig.show()

    def get_goal_on_index(self, name, population_id=-1):
        fig, ax = self.set_figure()
        table = self.goal_on_index(name=name, population_id=population_id)
        ax.plot(table[0], table[1])

        ax.set_xlabel(r'$k$', fontsize=self.label_size)
        ax.set_ylabel(r'$' + name + '$', fontsize=self.label_size)
        return [fig, ax]

    def goal_on_index_plot(self, name, population_id=-1):
        fig, ax = self.get_goal_on_index(name, population_id)
        fig.show()

    def get_parameter_on_index(self, name, population_id=-1):
        fig, ax = self.set_figure()
        table = self.parameter_on_index(name=name, population_id=population_id)
        ax.plot(table[0], table[1])

        ax.set_xlabel(r'$k$', fontsize=self.label_size)
        ax.set_ylabel(r'$' + name + '$', fontsize=self.label_size)
        return [fig, ax]

    def parameter_on_index_plot(self, name, population_id=-1):
        fig, ax = self.get_parameter_on_index(name, population_id)
        fig.show()

    def get_goal_on_parameter(self, parameter_name, goal_name, population_id=-1, sorted=False):
        fig, ax = self.set_figure()
        table = self.goal_on_parameter(parameter_name, goal_name, population_id, sorted)
        ax.plot(table[0], table[1], 'o')

        ax.set_xlabel(r'$' + parameter_name + '$', fontsize=self.label_size)
        ax.set_ylabel(r'$' + goal_name + '$', fontsize=self.label_size)
        return [fig, ax]

    def goal_on_parameter_plot(self, parameter_name, goal_name, population_id=-1, sorted=False):
        fig, ax = self.get_goal_on_parameter(parameter_name, goal_name, population_id, sorted)
        fig.show()

    def get_parameter_on_parameter(self, parameter_1, parameter_2, population_id=-1, sorted=False):
        fig, ax = self.set_figure()
        table = self.parameter_on_parameter(parameter_1, parameter_2, population_id, sorted)
        ax.plot(table[0], table[1], 'o')

        ax.set_xlabel(r'$' + parameter_1 + '$', fontsize=self.label_size)
        ax.set_ylabel(r'$' + parameter_2 + '$', fontsize=self.label_size)
        return [fig, ax]

    def parameter_on_parameter_plot(self, parameter_1, parameter_2, population_id=-1, sorted=False):
        fig, ax = self.get_parameter_on_parameter(parameter_1, parameter_2, population_id, sorted)
        fig.show()

    def goal_histogram_plot(self, name, population_id=-1):
        fig, ax = self.set_figure()
        table = self.goal_on_index(name=name, population_id=population_id)
        n, bins, patches = ax.hist(x=table[1], bins='auto', color='#0504aa',
                                    alpha=0.7, rwidth=0.85)
        fig.show()

    def get_mean_confidence_interval(self, name):
        index = self.goal_index(name)
        return 



import unittest

from ..problem import Problem
from ..algorithm_genetic import NSGAII
from ..results import Results
from random import random


class ParabolicProblem(Problem):
    """ Describe simple one objective optimization problem. """

    def set(self):
        self.name = "LocalPythonProblem"
        self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [-5, 5], 'precision': 1e-1},
                           {'name': 'x_2', 'initial_value': 2.5, 'bounds': [-5, 5], 'precision': 1e-1}]
        self.costs = [{'name': 'F_1'}]

    def evaluate(self, individual):
        x = individual.vector
        result = x[0]**2 + x[1]**2
        if random() > 0.9:
            raise TimeoutError
        return [result]


class TestFailedCalculation(unittest.TestCase):
    """ Tests that the nsga - ii can find the global optimum. """

    def test_local_problem(self, population_number=20):

        problem = ParabolicProblem()
        algorithm = NSGAII(problem)
        algorithm.options['max_population_number'] = population_number
        algorithm.options['max_population_size'] = 40
        algorithm.options['max_processes'] = 1
        algorithm.run()

        b = Results(problem)
        optimum = b.find_optimum('F_1')  # Takes last cost function
        self.assertAlmostEqual(optimum.costs[0], 0., 2)


if __name__ == '__main__':
    unittest.main()



import math
import unittest

from ..problem import Problem
from ..results import Results
from ..surrogate_smt import SurrogateModelSMT
from ..operators import LHSGenerator
from ..algorithm_sweep import SweepAlgorithm
from ..algorithm_nlopt import NLopt, LN_BOBYQA
from ..algorithm_genetic import NSGAII

import numpy as np
from smt.surrogate_models import RBF, MGP


class ProblemBranin(Problem):
    """
    Describe simple one objective optimization problem.
    x_min =  (-pi , 12.275), (pi , 2.275), (9.42478, 2.475), f(x) = 0.397887.
    """

    def set(self):
        self.parameters = [{'name': 'x_1', 'initial_value': 3, 'bounds': [-5, 10]},
                           {'name': 'x_2', 'initial_value': 10, 'bounds': [-5, 15]}]
        self.costs = [{'name': 'F'}]

    def evaluate(self, individual):
        x = individual.vector

        # a = 1
        # b = 5.1 / (4 * math.pi**2)
        # c = 5 / math.pi
        # r = 6
        # s = 10
        # t = 1 / (8 * math.pi)
        #
        # term1 = - a * (x[1] - b * x[0]**2 + c * x[0] - r)**2
        # term2 = - s * (1 - t) * math.cos(x[0])
        #
        # y = - term1 + term2 + s

        y = (x[0] + 2 * x[1] - 7) ** 2 + (2 * x[0] + x[1] - 5) ** 2

        return [y]

    def predict(self, individual):
        sigma_MGP, sigma_KRG = self.surrogate.predict_variances(individual.vector, True)

        if sigma_MGP < 1e-6:
            self.surrogate.train_step *= 1.3

            value_problem = self.evaluate(individual)
            value_surrogate = self.surrogate.predict(individual.vector)
            # print(individual.vector, value_problem, value_surrogate, sigma_MGP, sigma_KRG)

            return value_surrogate

        return None


class ProblemSphere1D(Problem):
    """
    Describe simple one objective optimization problem.
    """

    def set(self):
        self.parameters = [{'name': 'x_1', 'initial_value': 5, 'bounds': [-10, 10]}]
        self.costs = [{'name': 'F'}]

    def evaluate(self, individual):
        x = individual.vector

        y = x[0]**2

        return [y]

    def predict(self, individual):
        sigma_MGP, sigma_KRG = self.surrogate.predict_variances(individual.vector, True)

        if sigma_MGP < 1.00:
            # value_problem = self.evaluate(individual)
            value_surrogate = self.surrogate.predict(individual.vector)
            # print(individual.vector, value_problem, value_surrogate, sigma_MGP, sigma_KRG)

            return value_surrogate

        return None


class TestSurrogateFunction(unittest.TestCase):
    def test_function(self):
        problem = ProblemBranin()
        # problem = ProblemSphere1D()

        # surrogate
        problem.surrogate = SurrogateModelSMT(problem)
        # set custom regressor
        # problem.surrogate.regressor = RBF(d0=5, print_prediction=False)
        problem.surrogate.regressor = MGP(theta0=[1e-2], n_comp=2, print_prediction=False)

        # DoE - Latin - Hypercube
        gen = LHSGenerator(parameters=problem.parameters)
        gen.init(number=100)

        algorithm_sweep = SweepAlgorithm(problem, generator=gen)
        algorithm_sweep.run()

        # train model
        problem.surrogate.train()

        # set train step
        problem.surrogate.train_step = 50

        # run optimization
        # algorithm = NLopt(problem)
        # algorithm.options['verbose_level'] = 0
        # algorithm.options['algorithm'] = LN_BOBYQA
        # algorithm.options['xtol_abs'] = 1e-6
        # algorithm.options['xtol_rel'] = 1e-3
        # algorithm.options['ftol_rel'] = 1e-3
        # algorithm.options['ftol_abs'] = 1e-6
        # algorithm.options['n_iterations'] = 50
        # algorithm.run()

        algorithm = NSGAII(problem)
        algorithm.options['max_population_number'] = 40
        algorithm.options['max_population_size'] = 10
        algorithm.options['max_processes'] = 10
        algorithm.run()

        b = Results(problem)
        optimum = b.find_optimum('F_1')  # Takes last cost function
        # print(optimum.vector, optimum.costs)
        self.assertGreater(optimum.vector[0], 0.9)
        self.assertLess(optimum.vector[0], 1.1)
        self.assertGreater(optimum.vector[1], 2.8)
        self.assertLess(optimum.vector[1], 3.2)
        self.assertLess(optimum.costs[0], 0.05)
        self.assertGreater(problem.surrogate.predict_counter, 100)

        problem.logger.info("surrogate: predict / eval counter: {0:5.0f} / {1:5.0f}, total: {2:5.0f}".format(
            problem.surrogate.predict_counter,
            problem.surrogate.eval_counter,
            problem.surrogate.predict_counter + problem.surrogate.eval_counter))



from .problem import Problem
from .algorithm import Algorithm
from .config import artap_root

import time
import numpy as np

import os

import sys
sys.path.append(artap_root + os.sep + "lib" + os.sep)
import bayesopt

from multiprocessing import Process, Pipe, Queue, Manager
# from multiprocessing.managers import BaseManager

_l_type = ['L_FIXED', 'L_EMPIRICAL', 'L_DISCRETE', 'L_MCMC', 'L_ERROR']
_sc_type = ['SC_MTL', 'SC_ML', 'SC_MAP', 'SC_LOOCV', 'SC_ERROR']
_surr_name = ["sGaussianProcess", "sGaussianProcessML", "sGaussianProcessNormal", "sStudentTProcessJef", "sStudentTProcessNIG"]


# Python module to get run BayesOpt library in a OO pattern.
# The objective module should inherit this one and override evaluateSample.
class BayesOptContinuous(object):

    # Let's define the vector.
    #
    # For different options: see vector.h and vector.cpp .
    # If a parameter is not defined, it will be automatically set
    # to a default value.
    def __init__(self, n_dim):
        ## Library vector
        self.params = {}
        ## n dimensions
        self.n_dim = n_dim
        ## Lower bounds
        self.lb = np.zeros((self.n_dim,))
        ## Upper bounds
        self.ub = np.ones((self.n_dim,))

    @property
    def parameters(self):
        return self.params

    @parameters.setter
    def parameters(self, params):
        self.params = params

    @property
    def lower_bound(self):
        return self.lb

    @lower_bound.setter
    def lower_bound(self, lb):
        self.lb = lb

    @property
    def upper_bound(self):
        return self.ub

    @upper_bound.setter
    def upper_bound(self, ub):
        self.ub = ub

    ## Function for testing.
    # It should be overriden.
    def evaluateSample(self, x_in):
        raise NotImplementedError("Please Implement this method")

    ## Main function. Starts the optimization process.
    def optimize(self):
        min_val, x_out, error = bayesopt.optimize(self.evaluateSample, self.n_dim,
                                                  self.lb, self.ub,
                                                  self.params)

        return min_val, x_out, error


class BayesOpt(Algorithm):
    """ BayesOpt algorithms """

    def __init__(self, problem: Problem, name="BayesOpt"):
        super().__init__(problem, name)
        self.problem = problem

        self.options.declare(name='l_type', default='L_EMPIRICAL', values=_l_type,
                             desc='Type of learning for the kernel params')
        self.options.declare(name='sc_type', default='SC_MAP', values=_sc_type,
                             desc='Type of learning for the kernel params')
        self.options.declare(name='n_iterations', default=50, lower=1,
                             desc='Maximum BayesOpt evaluations')
        self.options.declare(name='init_method', default=1,
                             desc='Init method') # 1-LHS, 2-Sobol
        self.options.declare(name='n_init_samples', default=10, lower=1,
                             desc='Number of samples before optimization')
        self.options.declare(name='n_iter_relearn', default=10, lower=1,
                             desc='Number of samples before relearn kernel')
        self.options.declare(name='surr_name', default='sGaussianProcessML', values=_surr_name,
                             desc='Name of the surrogate function')
        self.options.declare(name='surr_noise', default=1e-10, lower=0.0,
                             desc='Variance of observation noise')


class BayesOptClassSerial(BayesOptContinuous):
    def __init__(self, algorithm):
        n = len(algorithm.problem.parameters)
        super().__init__(n)

        # algorithm
        self.algorithm = algorithm

        # Size design variables.
        self.lb = np.empty((n,))
        self.ub = np.empty((n,))
        self.params = {}

    def evaluateSample(self, x):
        return self.algorithm.evaluator.evaluate_scalar(x)


class BayesOptSerial(BayesOpt):
    """ BayesOpt algorithms """

    def __init__(self, problem: Problem, name="BayesOpt"):
        super().__init__(problem, name)

        self.bo = BayesOptClassSerial(self)

    def run(self):
        # Figure out bounds vectors.
        i = 0
        for parameter in self.problem.parameters:
            bounds = parameter['bounds']

            self.bo.lb[i] = bounds[0]
            self.bo.ub[i] = bounds[1]

            i += 1

        # set bayesopt
        self.bo.params['n_iterations'] = self.options['n_iterations']
        self.bo.params['n_init_samples'] = self.options['n_init_samples']
        self.bo.params['n_iter_relearn'] = self.options['n_iter_relearn']
        self.bo.params['surr_name'] = self.options['surr_name']
        self.bo.params['surr_noise'] = self.options['surr_noise']
        self.bo.params['init_method'] = self.options['init_method']
        self.bo.params['l_type'] = self.options['l_type']
        self.bo.params['sc_type'] = self.options['sc_type']
        self.bo.params['verbose_level'] = self.options['verbose_level']

        t_s = time.time()

        self.problem.logger.info("BayesOpt: surr_name{}".format(self.options['surr_name']))
        mvalue, x_out, error = self.bo.optimize()

        t = time.time() - t_s
        self.problem.logger.info("BayesOpt: elapsed time: {} s".format(t))

        # sync changed individual informations
        self.problem.data_store.sync_all()

        if error != 0:
            print('Optimization FAILED.')
            print("Error", error)
            print('-' * 35)
        else:
            pass
            # print('Optimization Complete, %f seconds' % (clock() - start))
            # print("Result", x_out, mvalue)
            # print('-' * 35)


class BayesOptClassParallel(Process, BayesOptContinuous):
    def __init__(self, pipe, algorithm):
        n = len(algorithm.problem.parameters)
        Process.__init__(self)
        BayesOptContinuous.__init__(self, n)

        # algorithm
        self.algorithm = algorithm

        # output
        self.mvalue = -1.0
        self.x_out = -1.0
        self.error = 0

        self.pipe = pipe

        # Size design variables.
        self.lb = np.empty((n,))
        self.ub = np.empty((n,))
        self.params = {}

    def run(self):
        mvalue, x_out, error = self.optimize()
        self.pipe.send('STOP')

        # set output values
        self.mvalue = mvalue
        self.x_out = x_out
        self.error = error

        # output
        print("output")
        print(self.mvalue)
        print(self.x_out)
        print(self.error)

    def evaluateSample(self, x):
        self.pipe.send(x)
        result = self.pipe.recv()
        return result


class BayesOptParallel(BayesOpt):
    """ BayesOpt algorithms """

    def __init__(self, problem: Problem, name="BayesOpt"):
        super().__init__(problem, name)

        self.pipe_par, self.pipe_child = Pipe()
        self.bo = BayesOptClassParallel(self.pipe_child, self)

    def worker(self, pipe):
        x = None
        while True:
            x = pipe.recv()
            if str(x) == 'STOP':
                break

            result = self.bo.job.evaluate_scalar(x)
            pipe.send(result)

    def run(self):
        # Figure out bounds vectors.
        i = 0
        for parameter in self.problem.parameters:
            bounds = parameter['bounds']

            self.bo.lb[i] = bounds[0]
            self.bo.ub[i] = bounds[1]

            i += 1

        # set bayesopt
        self.bo.params['n_iterations'] = self.options['n_iterations']
        self.bo.params['n_init_samples'] = self.options['n_init_samples']
        self.bo.params['n_iter_relearn'] = self.options['n_iter_relearn']
        self.bo.params['surr_name'] = self.options['surr_name']
        self.bo.params['surr_noise'] = self.options['surr_noise']
        self.bo.params['init_method'] = self.options['init_method']
        self.bo.params['l_type'] = self.options['l_type']
        self.bo.params['sc_type'] = self.options['sc_type']
        self.bo.params['verbose_level'] = self.options['verbose_level']

        # process = Process(target=self.worker, args=(self.pipe_par, self.problem, ))
        process = Process(target=self.worker, args=(self.pipe_par, ))

        self.bo.start()
        process.start()

        self.bo.join()
        process.join()

        print(self.bo.mvalue)
        print(self.bo.x_out)
        print(self.bo.error)
        print()
        print(self.problem.data_store, len(self.problem.populations[-1].individuals))


        # self.result = self.mvalue
        """
        if self.bo.error != 0:
            print('Optimization FAILED.')
            print("Error", self.bo.error)
            print('-' * 35)
        else:
            print('Optimization Complete, %f seconds' % (clock() - start))
            print("Result", self.bo.x_out, self.bo.mvalue)
            print('-' * 35)
            
        """